#!/usr/bin/env python
# -*- coding: utf-8 -*-

import torch
from torch.utils.data import Dataset
import pywt
import librosa

from scipy.signal import butter, lfilter
from helper_code import *

murmur_classes = ['Present', 'Unknown', 'Absent']
num_murmur_classes = len(murmur_classes)
outcome_classes = ['Abnormal', 'Normal']
num_outcome_classes = len(outcome_classes)

"""
Updated on 2022/08/6 by pan xia
"""


class PCG_Dataset(Dataset):
    def __init__(self, features, demographic, labels):
        self.x_data = features
        self.d_data = demographic
        self.y_data = labels
        self.len = len(labels)

    def __getitem__(self, index):
        x = torch.from_numpy(self.x_data[index]).float()
        d = torch.from_numpy(self.d_data[index]).float()
        # y = torch.from_numpy(self.y_data[index]).float()
        y = torch.tensor(self.y_data[index]).float()
        return x, d, y

    def __len__(self):
        return self.len


################################################################################
# data preprocess
################################################################################
def preprocess(data_folder, patient_files, verbose):
    inner_all_features = []
    inner_all_demographic_feature = []
    all_true_murmurs = []
    all_true_outcomes = []

    recording_length = []
    for i in range(len(patient_files)):
        if verbose >= 1:
            print('    {}/{}...'.format(i+1, len(patient_files)))

        # Load the current patient data and recordings.
        current_patient_data = load_patient_data(patient_files[i])
        current_frequencies = get_frequency(current_patient_data)
        current_recordings = load_recordings(data_folder, current_patient_data)
        for index in range(len(current_recordings)):
             recording_length.append(len(current_recordings[index])/current_frequencies)

        # Extract features.
        single_features, demographic_feature = get_pcg_features(current_patient_data, current_recordings, current_frequencies)
        inner_all_features.append(single_features)
        inner_all_demographic_feature.append(demographic_feature)

        # Extract murmur labels and use one-hot encoding.
        current_murmur = np.zeros(num_murmur_classes, dtype=int)
        murmur = get_murmur(current_patient_data)
        if murmur in murmur_classes:
            j = murmur_classes.index(murmur)
            current_murmur[j] = 1
        all_true_murmurs.append(current_murmur)

        # Extract outcome labels and use one-hot encoding.
        current_outcome = np.zeros(num_outcome_classes, dtype=int)
        outcome = get_outcome(current_patient_data)
        if outcome in outcome_classes:
            j = outcome_classes.index(outcome)
            current_outcome[j] = 1
        all_true_outcomes.append(current_outcome)

    return inner_all_features, inner_all_demographic_feature, all_true_murmurs, all_true_outcomes


################################################################################
# Extract features from the data.
################################################################################
def get_pcg_features(data, recordings, sample_Fs):
    # Extract the age group and replace with the (approximate) number of months for the middle of the age group.
    age_group = get_age(data)

    if compare_strings(age_group, 'Neonate'):
        age = 0.5
    elif compare_strings(age_group, 'Infant'):
        age = 6
    elif compare_strings(age_group, 'Child'):
        age = 6 * 12
    elif compare_strings(age_group, 'Adolescent'):
        age = 15 * 12
    elif compare_strings(age_group, 'Young Adult'):
        age = 20 * 12
    else:
        age = float('nan')

    # Extract sex. Use one-hot encoding.
    sex = get_sex(data)

    sex_female_features = 0
    sex_male_features = 0
    if compare_strings(sex, 'Female'):
        sex_female_features = 1
    elif compare_strings(sex, 'Male'):
        sex_male_features = 1

    # Extract height and weight.
    height = get_height(data)
    weight = get_weight(data)

    # Extract pregnancy status.
    is_pregnant = get_pregnancy_status(data)

    demographic_feature = [age, sex_female_features, sex_male_features, height, weight, is_pregnant]

    # Extract recording locations and data. Identify when a location is present, and compute the mean, variance, and skewness of
    # each recording. If there are multiple recordings for one location, then extract features from the last recording.
    front_noise = 5  # 去除前段噪声
    behind_noise = 3  # 去除后段噪声
    recording_average_length = 15  # 数据平均长度为23秒
    targetFreq = 1000  # Down-sampling to 500Hz
    locations = get_locations(data)

    recording_locations = ['AV', 'MV', 'PV', 'TV', 'PhC']
    recording_features = np.zeros((len(recording_locations), recording_average_length * targetFreq), dtype=float)
    if len(locations) == len(recordings):
        for i in range(len(locations)):
            for j in range(len(recording_locations)):
                if compare_strings(locations[i], recording_locations[j]) and np.size(recordings[i]) > 0:
                    down_filtered_pcg_lead = downsampling(recordings[i], sample_Fs, targetFreq)

                    truncated_pcg = down_filtered_pcg_lead[front_noise * targetFreq: (len(down_filtered_pcg_lead) - behind_noise * targetFreq)]
                    res = data_aligning(truncated_pcg[: recording_average_length * targetFreq], recording_average_length * targetFreq)  # 10s
                    recording_features[j] = res
                    recording_features[j] = channel_uniform(recording_features[j])
                    # CWT_Spectrum = LogMelExtractor(recording_features[j], targetFreq, log=True, snv=False)

    recording_features = recording_features.flatten()
    # print(recording_features.shape)

    # MFCC特征提取
    # mfccs = librosa.feature.mfcc(y=recording_features, sr=targetFreq, n_mfcc=24)
    # print(mfccs.shape)

    # 连续小波变换时频图
    # CWT_Spectrum = Time_Frequency_Spectrum(recording_features, targetFreq)
    # print(CWT_Spectrum.shape)

    # 2D梅尔频谱特征
    Mel_Spectrum = get_mel_spectrum(recording_features, targetFreq, log=True, snv=False)

    return Mel_Spectrum, demographic_feature


def channel_uniform(recording_features):
    # 归一化
    recording_features = np.array(recording_features, dtype='float32')  # 将64改成32，节省内存
    theMean = np.mean(recording_features).tolist()
    theStd = np.std(recording_features).tolist()
    uniform_recording_features = (recording_features - theMean) / theStd

    uniform_recording_features[np.isnan(uniform_recording_features)] = 0.01
    uniform_recording_features[np.isinf(uniform_recording_features)] = 0.01

    return uniform_recording_features


################################################################################
# data uniform
################################################################################
def Time_Frequency_Spectrum(signal, Fs):
    Wavename = 'morl'
    Totalscale = 256
    fc = pywt.central_frequency(Wavename)
    Cparam = 2 * fc * Totalscale
    Scales = Cparam / np.arange(Totalscale, 1, -1)
    [Cwtmatr, frequencies] = pywt.cwt(signal, Scales, Wavename, 1.0 / Fs)
    # plt.figure(3)
    # plt.contourf(abs(Cwtmatr))
    # plt.show()

    return abs(Cwtmatr)


################################################################################
# log mel spectrum
################################################################################
def get_mel_spectrum(data, sr, mel_bins=32, hoplen=3, winlen=25, log=True, snv=False):
    # define parameters,
    MEL_ARGS = {
        'n_mels': mel_bins,
        'sr': sr,
        'hop_length': int(sr * hoplen / 1000),
        'win_length': int(sr * winlen / 1000)
    }

    mel_spectrogram = librosa.feature.melspectrogram(data, **MEL_ARGS)
    if log:
        mel_spectrogram = np.log(mel_spectrogram + 1E-8)
    if snv:
        mel_spectrogram = standard_normal_variate(mel_spectrogram)

    mel_spectrogram = channel_uniform(mel_spectrogram)

    return mel_spectrogram


################################################################################
# normalize mel spectrum
################################################################################
def standard_normal_variate(data):
    mean = np.mean(data)
    std = np.std(data)

    return (data - mean) / std


################################################################################
# convert data to array
################################################################################
def convert_array(features, demographics, murmur_labels, outcome_labels):

    features = np.array(features, dtype='float32')
    demographics = np.array(demographics, dtype='float32')
    murmur_labels = np.array(murmur_labels)
    outcome_labels = np.array(outcome_labels)

    return features, demographics, murmur_labels, outcome_labels


################################################################################
# Down sampling
################################################################################
def downsampling(data, sample_fs, targetFreq):
    step = sample_fs // targetFreq
    resample_data = [data[i] for i in range(len(data)) if i % step == 0]
    return resample_data


################################################################################
# data aligning
################################################################################
def data_aligning(data, window):
    if len(data) == window:
        return data
    elif len(data) < window:
        data += [0 for i in range(window - len(data))]
        return data
    else:
        raise Exception("Error in dataLenCheck")
    # 防止数据片段小于window