#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on 2022/07/15
@author: pan xia
CNN + Transformer model
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.model_zoo as model_zoo
import math
from torch.autograd import Variable


# weighted focal loss
class custom_focal_loss(nn.Module):
    def __init__(self, alpha=0.5, gamma=2, lamda=1, num_classes=0, size_average=True):
        super(custom_focal_loss, self).__init__()
        self.size_average = size_average
        if isinstance(alpha, list):
            assert len(alpha) == num_classes
            self.alpha = torch.tensor(alpha)
        else:
            assert alpha < 1
            self.alpha = torch.zeros(num_classes)
            self.alpha[0] += alpha
            self.alpha[1:] += (1 - alpha)

        self.gamma = gamma
        if num_classes == 3:
            self.theta = torch.tensor([1, 0, -1])
        elif num_classes == 2:
            self.theta = torch.tensor([1, -1])

        self.lamda = lamda

    def forward(self, preds, labels):
        # assert preds.dim()==2 and labels.dim()==1
        preds = preds.view(-1, preds.size(-1))

        self.alpha = self.alpha.to(preds.device)
        self.theta = self.theta.to(preds.device)

        preds_softmax = F.softmax(preds, dim=1)
        preds_logsoft = torch.log(preds_softmax)

        # RuntimeError: gather_out_cuda(): Expected dtype int64 for index
        labels = labels.type(torch.int64)

        # focal_loss func, Loss = -α(1-yi)**γ *ce_loss(xi,yi)
        preds_softmax = preds_softmax.gather(1, labels.view(-1, 1))  # 选出来label对应的softmax预测
        # print(preds_softmax)
        preds_logsoft = preds_logsoft.gather(1, labels.view(-1, 1))  # 选出来label对应的log_softmax预测
        # print(preds_logsoft)

        # 选出label对应的alpha权重
        alpha = self.alpha.gather(0, labels.view(-1))
        # print(alpha)

        # 选出label对应的theta
        theta = self.theta.gather(0, labels.view(-1))
        # print(labels)
        # print(theta)

        # 为gamma加权
        weighted_gamma = torch.mul(self.gamma, torch.exp(self.lamda * theta))
        # print(weighted_gamma)

        # torch.pow((1-preds_softmax), self.gamma) 为focal loss中 (1-pt)**γ
        loss = -torch.mul(torch.pow((1 - preds_softmax), weighted_gamma), preds_logsoft)

        # 类别加权
        loss = torch.mul(alpha, loss.t())
        if self.size_average:
            loss = loss.mean()
        else:
            loss = loss.sum()
        return loss


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, in_channels, out_channels, stride=1, downsample=None):
        super(BasicBlock, self).__init__()
        self.conv1 = nn.Conv2d(in_channels, out_channels, (3, 11), stride, padding=(1, 5), bias=False)
        self.bn1 = nn.BatchNorm2d(out_channels)
        self.conv2 = nn.Conv2d(out_channels, out_channels, (3, 11), 1, padding=(1, 5), bias=False)
        self.bn2 = nn.BatchNorm2d(out_channels)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample

    def forward(self, input):
        residual = input
        x = self.conv1(input)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.conv2(x)
        x = self.bn2(x)
        if self.downsample:
            residual = self.downsample(residual)
        x += residual
        x = self.relu(x)
        return x


class ResNet(nn.Module):
    # 128*3572
    def __init__(self, block, num_layer, n_classes, input_channels, dropout_fc):
        super(ResNet, self).__init__()
        self.in_channels = 64
        self.conv1 = nn.Conv2d(input_channels, 64, kernel_size=(3, 11), stride=(2, 2), padding=(1, 5), bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d((2, 2), stride=(1, 2), padding=1)
        self.layer1 = self._make_layer(block, 64, num_layer[0])
        self.layer2 = self._make_layer(block, 128, num_layer[1], 2)
        self.layer3 = self._make_layer(block, 128, num_layer[2], 2)
        self.layer4 = self._make_layer(block, 256, num_layer[3], 2)
        self.layer5 = self._make_layer(block, 256, num_layer[4], 2)
        self.layer6 = self._make_layer(block, 512, num_layer[5], 2)
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        # self.fc = nn.Linear(block.expansion * 512, n_classes)
        self.fc = nn.Sequential(nn.Linear(block.expansion * 512, 128), nn.BatchNorm1d(128), nn.ReLU(), nn.Dropout(dropout_fc),
                                nn.Linear(128, n_classes))
        self.fc1 = nn.Linear(512, 64)
        self.relu1 = nn.ReLU()

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1.0)
                nn.init.constant_(m.bias, 0.0)

    def _make_layer(self, block, out_channels, num_block, stride=1):
        downsample = None
        if stride != 1 or self.in_channels != out_channels * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.in_channels, out_channels * block.expansion, 1, stride=stride, bias=False),
                nn.BatchNorm2d(out_channels * block.expansion)
            )
        layers = []
        layers.append(block(self.in_channels, out_channels, stride, downsample))
        self.in_channels = out_channels * block.expansion
        for _ in range(1, num_block):
            layers.append(block(self.in_channels, out_channels))
        return nn.Sequential(*layers)

    def forward(self, input):
        x = self.conv1(input)
        # print("conv1", x.size())
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)
        # print("maxpool", x.size())

        x = self.layer1(x)
        x = F.dropout(x, p=0.2, training=self.training)
        # print("layer1", x.size())
        x = self.layer2(x)
        x = F.dropout(x, p=0.2, training=self.training)
        # print("layer2", x.size())
        x = self.layer3(x)
        x = F.dropout(x, p=0.2, training=self.training)
        # print("layer3", x.size())
        x = self.layer4(x)
        x = F.dropout(x, p=0.2, training=self.training)
        # print("layer4", x.size())
        x = self.layer5(x)
        x = F.dropout(x, p=0.2, training=self.training)
        # print("layer5", x.size())
        x = self.layer6(x)
        x = F.dropout(x, p=0.2, training=self.training)
        # print("layer6", x.size())

        # x = self.avgpool(x)
        # x = x.flatten(start_dim=2)
        # print("flatten", x.size())
        # x = torch.mean(x, dim=2)
        x = F.avg_pool2d(x, kernel_size=(x.size(2), x.size(3)), stride=1)
        # print("GAP", x.size())
        z = x.view(x.size(0), -1)
        # print("view", x.size())

        x = self.relu1(self.fc1(z))
        y = self.fc(z)
        # print(x.size())
        return x, y


class MyModel(nn.Module):
    def __init__(self, in_channel, n_class, dropout_fc):
        super(MyModel, self).__init__()

        self.n_class = n_class
        self.dropout_fc = dropout_fc
        self.in_channel = in_channel

        self.resnet = ResNet(block=BasicBlock, num_layer=[2, 2, 2, 2, 2, 2], n_classes=n_class, input_channels=in_channel, dropout_fc=dropout_fc)

    def forward(self, input, demographic):
        # print("input", input.size())

        batch, mel_freq, mel_time = input.size()  # input size N, H, W
        # print(batch, mel_freq, mel_time)
        x = input.view(batch, self.in_channel, -1, mel_time)
        # print(x.size())

        x = self.resnet(x)
        # print("Resnet", x.size())

        return x
