#!/usr/bin/env python

################################################################################
#
# Import libraries and functions. You can change or remove them.
#
################################################################################
import os
import numpy as np
import torch
from scipy.io import loadmat
from torch.utils.data import DataLoader
from tqdm import tqdm
from sklearn.model_selection import train_test_split, KFold, StratifiedKFold
from sklearn.impute import SimpleImputer

from data_preprocess import *
from resnet import *
from helper_code import *
from center_loss import CenterLoss
from mixup import *


# 每个epoch训练函数
def train(epoch, model, optimizer, optimizer_centloss, train_loader, use_cuda, criterion, center_loss, alpha, weight_center):
    model.train()
    total_loss = 0
    n_entries = 0
    all_logits = []
    all_labels = []
    train_desc = "Epoch {:2d}: train - Loss: {:.6f}"
    train_bar = tqdm(initial=0, leave=True, total=len(train_loader), desc=train_desc.format(epoch, 0), position=0)
    for i, batch in enumerate(train_loader):
        data, demographic, target = batch
        if use_cuda:
            data = data.cuda()
            target = target.cuda()
            demographic = demographic.cuda()

        # generate mixed inputs, two one-hot label vectors and mixing coefficient
        inputs, targets_a, targets_b, lam = mixup_data(data, target, alpha, use_cuda)
        inputs, targets_a, targets_b = Variable(inputs), Variable(targets_a), Variable(targets_b)
        feature, outputs = model(inputs, demographic)

        loss_func = mixup_criterion(targets_a, targets_b, lam)
        loss_focal = loss_func(criterion, outputs)
        loss_center = weight_center * center_loss(feature, target.long())  # weight * center loss
        loss = loss_focal + loss_center
        optimizer.zero_grad()
        optimizer_centloss.zero_grad()
        loss.backward()
        optimizer.step()
        # by doing so, weight_center would not impact on the learning of centers
        for param in center_loss.parameters():
            param.grad.data *= (1. / weight_center)
        optimizer_centloss.step()

        batch_loss = loss.detach().cpu().numpy()
        total_loss += batch_loss
        logits_softmax = torch.softmax(outputs, dim=1)
        bs = target.size(0)
        n_entries += bs
        all_logits.append(logits_softmax.detach().cpu())

        # convert index label to one-hot label
        target = target.view(-1, 1).cpu()
        # print("index label", target)
        one_hot_target = torch.zeros(outputs.size(0), outputs.size(1)).scatter_(1, target.long(), 1)
        # print("one-hot label", one_hot_target)
        all_labels.append(one_hot_target.cpu())
        train_bar.desc = train_desc.format(epoch, total_loss / n_entries)
        train_bar.update(bs)
    train_bar.close()

    return total_loss / n_entries, torch.cat(all_labels), torch.cat(all_logits)


# 每个epoch验证函数
def validate(epoch, model, valid_loader, use_cuda, criterion, center_loss, weight_center):
    model.eval()
    total_loss = 0
    n_entries = 0
    all_outputs = []
    all_labels = []
    eval_desc = "Epoch {:2d}: valid - Loss: {:.6f}"
    eval_bar = tqdm(initial=0, leave=True, total=len(valid_loader),
                    desc=eval_desc.format(epoch, 0), position=0)
    for i, batch in enumerate(valid_loader):
        with torch.no_grad():
            data, demographic, target = batch
            if use_cuda:
                data = data.cuda()
                demographic = demographic.cuda()
                target = target.cuda()

            feature, output = model(data, demographic)
            loss_focal = criterion(output, target.long())
            loss_center = weight_center * center_loss(feature, target.long())  # weight * center loss
            batch_loss = loss_focal + loss_center
            total_loss += batch_loss.item()
            output_softmax = torch.softmax(output, dim=1)  # output.size = batch * num_class，故而dim=1
            all_outputs.append(output_softmax.detach().cpu())

            # convert index label to one-hot label
            target = target.view(-1, 1).cpu()
            # print(target)
            one_hot_target = torch.zeros(output.size(0), output.size(1)).scatter_(1, target.long(), 1)
            # print(one_hot_target)
            all_labels.append(one_hot_target.cpu())
            bs = data.size(0)
            n_entries += bs
            eval_bar.desc = eval_desc.format(epoch, total_loss / n_entries)
            eval_bar.update(bs)
    eval_bar.close()
    return total_loss / n_entries, torch.cat(all_labels), torch.cat(all_outputs)


################################################################################
#
# Required functions. Edit these functions to add your code, but do not change the arguments.
#
################################################################################
def train_challenge_model(data_folder, model_folder, verbose):
    # Find data files.
    if verbose >= 1:
        print('Finding data files...')

    # Find the patient data files.
    patient_files = find_patient_files(data_folder)
    num_patient_files = len(patient_files)

    if num_patient_files == 0:
        raise Exception('No data was provided.')

    # Create a folder for the model if it does not already exist.
    os.makedirs(model_folder, exist_ok=True)

    # Extract the features and labels.
    if verbose >= 1:
        print('Extracting features and labels from the Challenge data...')

    # **************************加载数据、解析特征 **************************
    totalFeatures, totalDemographic, totalMurmurs, totalOutcomes = preprocess(data_folder, patient_files, verbose)

    imputer = SimpleImputer().fit(totalDemographic)
    totalDemographic = imputer.transform(totalDemographic)

    # **************************数据归一化处理 **************************
    uniform_features, uniform_Demographic, uniform_Murmurs, uniform_Outcomes = convert_array(totalFeatures, totalDemographic, totalMurmurs, totalOutcomes)

    # Convert one-hot to index encoding
    uniform_Murmurs = uniform_Murmurs.argmax(1)
    uniform_Murmurs = np.array(uniform_Murmurs)
    uniform_Outcomes = uniform_Outcomes.argmax(1)
    uniform_Outcomes = np.array(uniform_Outcomes)

    print(uniform_features.shape, uniform_Murmurs.shape, uniform_Outcomes.shape)

    # ***************************Murmur模型5折交叉验证****************************
    # 网络超参数
    batch_size = 16
    learning_rate = 0.0001
    num_epochs = 40
    in_channel = 1  # input channel is 1
    dropout_fc = 0.1  # dropout value for feedforward output layers
    murmur_mixup_alpha = 0.1
    lamda_murmur = 2  # control the strength of weighted for maximizing weighted accuracy, suggested [0, 2)
    certerloss_feature_size = 64
    weight_center_loss = 0.1  # weight of center loss

    # ***************************Murmur模型****************************
    # 加载数据至Dataloader接口
    dataset_train = PCG_Dataset(uniform_features, uniform_Demographic, uniform_Murmurs)
    train_iter = DataLoader(dataset_train, batch_size=batch_size, shuffle=True, num_workers=0, drop_last=True)

    # 其他参数设置
    use_cuda = torch.cuda.is_available()
    loss_func = custom_focal_loss(lamda=lamda_murmur, num_classes=len(murmur_classes))

    # 实例化model
    murmur_model = MyModel(in_channel=in_channel, n_class=len(murmur_classes), dropout_fc=dropout_fc)
    for param in murmur_model.parameters():  # 使用uniform distribution初始化模型参数
        if param.dim() > 1:
            nn.init.xavier_uniform_(param)

    if use_cuda:
        murmur_model = murmur_model.cuda()
    optimizer = torch.optim.Adam(murmur_model.parameters(), lr=learning_rate, betas=(0.9, 0.98), eps=1e-9, weight_decay=0.0001)

    scheduler_lr = torch.optim.lr_scheduler.MultiStepLR(optimizer, milestones=[10, 20, 40], gamma=0.5, last_epoch=-1)

    center_loss = CenterLoss(num_classes=len(murmur_classes), feat_dim=certerloss_feature_size, use_gpu=use_cuda)
    optimizer_centloss = torch.optim.SGD(center_loss.parameters(), lr=0.5)

    # ***************************训练Murmur模型****************************
    for epoch in range(num_epochs):
        t_loss, t_labels, t_outputs = train(epoch + 1, murmur_model, optimizer, optimizer_centloss, train_iter,
                                            use_cuda, loss_func, center_loss, murmur_mixup_alpha, weight_center_loss)
        scheduler_lr.step()

    # 保存模型
    print('Saving model...')
    model_filename = 'Murmur_model.pt'
    torch.save(murmur_model, os.path.join(model_folder, model_filename))

    print('Murmur Done.')
    ################################################################################

    # ***************************Outcome模型5折交叉验证****************************
    # 网络超参数
    outcome_batch_size = 16
    outcome_learning_rate = 0.0001
    outcome_num_epochs = 40
    outcome_in_channel = 1  # input channel is 1
    outcome_dropout_fc = 0.3  # dropout value for feedforward output layers
    num_class_outcome = len(outcome_classes)
    outcome_mixup_alpha = 0.0001
    lamda_outcome = 0.5  # control the strength of weighted for maximizing weighted accuracy, suggested [0, 2)
    certerloss_feature_size = 64
    weight_center_loss = 1  # weight of center loss

    # ***************************Outcome模型****************************
    # 加载数据至Dataloader接口
    dataset_train = PCG_Dataset(uniform_features, uniform_Demographic, uniform_Outcomes)
    train_iter = DataLoader(dataset_train, batch_size=outcome_batch_size, shuffle=True, num_workers=0, drop_last=True)

    # 其他参数设置
    use_cuda = torch.cuda.is_available()
    loss_func = custom_focal_loss(lamda=lamda_outcome, num_classes=num_class_outcome)

    # 实例化model
    outcome_model = MyModel(in_channel=outcome_in_channel, n_class=num_class_outcome, dropout_fc=outcome_dropout_fc)
    for param in outcome_model.parameters():  # 使用uniform distribution初始化模型参数
        if param.dim() > 1:
            nn.init.xavier_uniform_(param)

    if use_cuda:
        outcome_model = outcome_model.cuda()
    optimizer = torch.optim.Adam(outcome_model.parameters(), lr=outcome_learning_rate, betas=(0.9, 0.98), eps=1e-9, weight_decay=0.0001)

    scheduler_lr = torch.optim.lr_scheduler.MultiStepLR(optimizer, milestones=[10, 20, 40], gamma=0.5, last_epoch=-1)

    center_loss = CenterLoss(num_classes=num_class_outcome, feat_dim=certerloss_feature_size, use_gpu=use_cuda)
    optimizer_centloss = torch.optim.SGD(center_loss.parameters(), lr=0.5)

    # ***************************训练Outcome模型****************************
    for epoch in range(outcome_num_epochs):
        t_loss, t_labels, t_outputs = train(epoch + 1, outcome_model, optimizer, optimizer_centloss, train_iter,
                                            use_cuda, loss_func, center_loss, outcome_mixup_alpha, weight_center_loss)
        scheduler_lr.step()

    # 保存模型
    print('Saving Outcome model...')
    model_filename = 'Outcome_model.pt'
    torch.save(outcome_model, os.path.join(model_folder, model_filename))

    print('Outcome Done.')
    print('All Done.')
    ################################################################################


# Load your trained model. This function is *required*. You should edit this function to add your code,
# but do *not* change the arguments of this function.
def load_challenge_model(model_folder, verbose):
    # 加载Murmur/Outcome模型
    model_filename_murmur = 'Murmur_model.pt'
    model_filename_outcome = 'Outcome_model.pt'
    root_murmur, _ = os.path.splitext(model_filename_murmur)
    root_outcome, _ = os.path.splitext(model_filename_outcome)

    dirs = os.listdir(model_folder)
    for file in dirs:
        root_file, _ = os.path.splitext(file)
        if root_murmur == root_file:
            if str(file).endswith(".pt"):
                Murmur_model = torch.load(os.path.join(model_folder, file))

        if root_outcome == root_file:
            if str(file).endswith(".pt"):
                Outcome_model = torch.load(os.path.join(model_folder, file))

    Murmur_classes = murmur_classes
    Outcome_classes = outcome_classes

    return [Murmur_classes, Murmur_model, Outcome_classes, Outcome_model]


# Run your trained model. This function is *required*. You should edit this function to add your code,
# but do *not* change the arguments of this function.
def run_challenge_model(model, patient_data, recordings, verbose):
    murmur_classes = model[0]  # murmur classes
    murmur_classifier = model[1]  # murmur classifier
    outcome_classes = model[2]  # outcome classes
    outcome_classifier = model[3]  # outcome classifier

    # Extract features.
    current_frequencies = get_frequency(patient_data)
    single_features, demographic_feature = get_pcg_features(patient_data, recordings, current_frequencies)

    single_features = np.array(single_features, dtype='float32')
    demographic_feature = np.array(demographic_feature, dtype='float32')

    # convert time-freq feature to tensor and move to gpu
    single_features = torch.from_numpy(single_features).float()
    single_features = single_features.reshape(1, single_features.size(0), single_features.size(1))
    single_features = single_features.to(next(murmur_classifier.parameters()).device)

    # convert demographic feature to tensor and move to gpu
    demographic_feature = torch.from_numpy(demographic_feature).float()
    demographic_feature = demographic_feature.reshape(1, demographic_feature.size(0))
    demographic_feature = demographic_feature.to(next(murmur_classifier.parameters()).device)

    # murmur output
    murmur_classifier.eval()
    _, murmur_output = murmur_classifier(single_features, demographic_feature)
    murmur_output = murmur_output.to('cpu').detach().numpy()
    murmur_output = np.array(murmur_output, dtype='float32')[:, 0:3]
    murmur_output_list = murmur_output.tolist()[0]

    # outcome output
    outcome_classifier.eval()
    _, outcome_output = outcome_classifier(single_features, demographic_feature)
    outcome_output = outcome_output.to('cpu').detach().numpy()
    outcome_output = np.array(outcome_output, dtype='float32')[:, 0:2]
    outcome_output_list = outcome_output.tolist()[0]

    murmur_pred_prob = torch.softmax(torch.tensor(murmur_output_list), dim=0)
    murmur_probabilities_list = murmur_pred_prob.tolist()
    # print("murmur_probabilities_list", murmur_probabilities_list)
    outcome_pred_prob = torch.softmax(torch.tensor(outcome_output_list), dim=0)
    outcome_probabilities_list = outcome_pred_prob.tolist()
    # print("outcome_probabilities_list", outcome_probabilities_list)

    # Choose label with highest probability.
    murmur_labels = np.zeros(len(murmur_classes), dtype=np.int_)
    if np.max(murmur_probabilities_list) >= 0.5:
        idx = np.argmax(murmur_probabilities_list)
        murmur_labels[idx] = 1

    outcome_labels = np.zeros(len(outcome_classes), dtype=np.int_)
    if np.max(outcome_probabilities_list) >= 0.55:
        idx = np.argmax(outcome_probabilities_list)
        outcome_labels[idx] = 1

    outcome_probabilities_list = [0, 0]
    # Concatenate classes, labels, and probabilities.
    classes = murmur_classes + outcome_classes
    pred_labels = np.concatenate((murmur_labels, outcome_labels))
    pred_probabilities = np.concatenate((murmur_probabilities_list, outcome_probabilities_list))

    return classes, pred_labels, pred_probabilities
